#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-26
#
# Licensed under Apache License, Version 2.0.
#

import typing

from . import assembly_factory as pf
from . import logging as pc_logging
from .utils import resolve_resource_path


class AssemblyFactoryAlias(pf.AssemblyFactory):
    source_assembly_name: str
    source_project_name: typing.Optional[str]
    source: str

    def __init__(self, ctx, project, assembly_config):
        with pc_logging.Action(
            "InitAlias", project.name, assembly_config["name"]
        ):
            super().__init__(ctx, project, assembly_config)
            # Complement the config object here if necessary
            self._create(assembly_config)

            if "source" in assembly_config:
                self.source_assembly_name = assembly_config["source"]
            else:
                self.source_assembly_name = assembly_config["name"]
                if not "project" in assembly_config:
                    raise Exception(
                        "Alias needs either the source part name or the source project name"
                    )

            if "project" in assembly_config:
                self.source_project_name = assembly_config["project"]
                if (
                    self.source_project_name == "this"
                    or self.source_project_name == ""
                ):
                    self.source_project_name = project.name
            else:
                if ":" in self.source_assembly_name:
                    self.source_project_name, self.source_assembly_name = (
                        resolve_resource_path(
                            project.name,
                            self.source_assembly_name,
                        )
                    )
                else:
                    self.source_project_name = project.name
            self.source = (
                self.source_project_name + ":" + self.source_assembly_name
            )

            if self.source_project_name == project.name:
                self.assembly.desc = "Alias to %s" % self.source_assembly_name
            else:
                self.assembly.desc = "Alias to %s from %s" % (
                    self.source_assembly_name,
                    self.source_project_name,
                )

            pc_logging.debug("Initializing an alias to %s" % self.source)

    def instantiate(self, assembly):
        with pc_logging.Action("Alias", assembly.project_name, assembly.name):

            source = self.ctx._get_assembly(self.source)
            shape = source.shape
            if not shape is None:
                assembly.shape = shape
                return

            self.ctx.stats_assemblies_instantiated += 1

            source.instantiate(assembly)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.
#

from . import part_factory_file as pff
from .runtime_python import PythonRuntime


class PartFactoryPython(pff.PartFactoryFile):
    runtime: PythonRuntime

    def __init__(
        self, ctx, source_project, target_project, part_config, can_create=False
    ):
        super().__init__(
            ctx,
            source_project,
            target_project,
            part_config,
            extension=".py",
            can_create=can_create,
        )

        self.runtime = self.ctx.get_python_runtime(self.project.python_version)

    async def prepare_python(self):
        """
        This method is called by child classes
        to prepare the Python environment
        before instantiating the part.
        """
        await self.runtime.prepare_for_package(self.project)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

import base64
import mimetypes
from pathlib import Path
import threading
from typing import Any

import openai as openai_genai

from . import logging as pc_logging
from .user_config import user_config

lock = threading.Lock()
OPENAI_API_KEY = None
openai_client = None

model_tokens = {
    "gpt-3": 4000,
    "gpt-3.5-turbo": 4096,
    "gpt-4": 8000,  # 32600,
    "gpt-4-vision-preview": 8192,
}


def openai_once():
    global OPENAI_API_KEY, openai_client

    with lock:
        latest_key = user_config.openai_api_key
        if latest_key != OPENAI_API_KEY:
            OPENAI_API_KEY = latest_key
            if not OPENAI_API_KEY is None:
                openai_genai.api_key = OPENAI_API_KEY
                openai_client = openai_genai.OpenAI(api_key=OPENAI_API_KEY)
                return True

        if OPENAI_API_KEY is None:
            error = "OpenAI API key is not set"
            pc_logging.error(error)
            return False

    return True


class AiOpenAI:
    def generate_openai(
        self,
        model: str,
        prompt: str,
        image_filenames: list[str] = [],
        config: dict[str, Any] = {},
        options_num: int = 1,
    ):
        if not openai_once():
            return None

        if "tokens" in config:
            tokens = config["tokens"]
        else:
            tokens = model_tokens[model]

        if "top_p" in config:
            top_p = config["top_p"]
        else:
            top_p = None

        content = [
            {"type": "text", "text": prompt},
            *list(
                map(
                    lambda f: {
                        "type": "image_url",
                        "image_url": {
                            "url": "data:%s;base64,%s"
                            % (
                                mimetypes.guess_type(f, False)[0],
                                base64.b64encode(Path(f).read_bytes()).decode(),
                            ),
                            "detail": "high",
                        },
                    },
                    image_filenames,
                )
            ),
        ]

        cc = openai_client.chat.completions.create(
            messages=[
                {"role": "system", "content": "You are a mechanical engineer"},
                {"role": "user", "content": content},
            ],
            stream=False,
            n=options_num,
            max_tokens=tokens,
            top_p=top_p,
            model=model,
        )

        products = []
        try:
            for choice in cc.choices:
                if hasattr(choice, "role") and choice.role == "system":
                    continue

                script = choice.message.content

                products.append(script)
        except Exception as e:
            pc_logging.exception(e)

        return products



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

import mimetypes
from pathlib import Path
import threading
import time
from typing import Any

import google.generativeai as google_genai
import google.api_core.exceptions

from . import logging as pc_logging
from .user_config import user_config

lock = threading.Lock()
GOOGLE_API_KEY = None

model_tokens = {
    "gemini-pro": 32600,
    "gemini-pro-vision": 16300,
}


def google_once():
    global GOOGLE_API_KEY

    with lock:
        latest_key = user_config.google_api_key
        if latest_key != GOOGLE_API_KEY:
            GOOGLE_API_KEY = latest_key
            if not GOOGLE_API_KEY is None:
                google_genai.configure(api_key=GOOGLE_API_KEY)
                return True

        if GOOGLE_API_KEY is None:
            error = "Google API key is not set"
            pc_logging.error(error)
            return False

    return True


class AiGoogle:
    def generate_google(
        self,
        model: str,
        prompt: str,
        image_filenames: list[str] = [],
        config: dict[str, Any] = {},
        options_num: int = 1,
    ):
        if not google_once():
            return None

        if "tokens" in config:
            tokens = config["tokens"]
        else:
            tokens = model_tokens[model]

        images = list(
            map(
                lambda f: {
                    "mime_type": mimetypes.guess_type(f, False)[0],
                    "data": Path(f).read_bytes(),
                },
                image_filenames,
            )
        )
        contents = [prompt, *images]

        client = google_genai.GenerativeModel(
            model,
            generation_config={
                # "candidate_count": options_num,
                "max_output_tokens": tokens,
                # "temperature": 0.97,
            },
        )
        candidates = []
        for _ in range(options_num):
            retry = True
            while retry == True:
                retry = False
                try:
                    response = client.generate_content(
                        contents,
                    )
                except google.api_core.exceptions.ResourceExhausted as e:
                    pc_logging.exception(e)
                    retry = True
                    time.sleep(60)
                except google.api_core.exceptions.InternalServerError as e:
                    pc_logging.exception(e)
                    retry = True
                    time.sleep(1)

            if response is None:
                error = "%s: Failed to generate content" % self.name
                pc_logging.error(error)
                continue
            candidates.extend(response.candidates)

        products = []
        try:
            for candidate in candidates:
                product = ""
                for part in candidate.content.parts:
                    if "text" in part:
                        product += part.text + "\n"

                products.append(product)
        except Exception as e:
            pc_logging.exception(e)

        return products



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

import base64
import os
import pickle
import sys

from . import part_factory_python as pfp
from . import wrapper
from . import logging as pc_logging

sys.path.append(os.path.join(os.path.dirname(__file__), "wrappers"))
from cq_serialize import register as register_cq_helper


class PartFactoryCadquery(pfp.PartFactoryPython):
    def __init__(
        self, ctx, source_project, target_project, part_config, can_create=False
    ):
        with pc_logging.Action(
            "InitCadQuery", target_project.name, part_config["name"]
        ):
            super().__init__(
                ctx,
                source_project,
                target_project,
                part_config,
                can_create=can_create,
            )
            # Complement the config object here if necessary
            self._create(part_config)

    async def instantiate(self, part):
        with pc_logging.Action("CadQuery", part.project_name, part.name):
            if not os.path.exists(part.path) or os.path.getsize(part.path) == 0:
                pc_logging.error(
                    "CadQuery script is empty or does not exist: %s" % part.path
                )
                return None

            # Finish initialization of PythonRuntime
            # which was too expensive to do in the constructor
            await self.prepare_python()

            # Get the path to the wrapper script
            # which needs to be executed
            wrapper_path = wrapper.get("cadquery.py")

            # Build the request
            request = {"build_parameters": {}}
            if "parameters" in self.part_config:
                for param_name, param in self.part_config["parameters"].items():
                    request["build_parameters"][param_name] = param["default"]

            # Serialize the request
            register_cq_helper()
            picklestring = pickle.dumps(request)
            request_serialized = base64.b64encode(picklestring).decode()

            await self.runtime.ensure("cadquery")
            response_serialized, errors = await self.runtime.run(
                [
                    wrapper_path,
                    os.path.abspath(part.path),
                    os.path.abspath(self.project.config_dir),
                ],
                request_serialized,
            )
            if len(errors) > 0:
                error_lines = errors.split("\n")
                for error_line in error_lines:
                    part.error("%s: %s" % (part.name, error_line))

            try:
                response = base64.b64decode(response_serialized)
                register_cq_helper()
                result = pickle.loads(response)
            except Exception as e:
                part.error(
                    "Exception while deserializing %s: %s" % (part.name, e)
                )
                return None

            if not result["success"]:
                part.error("%s: %s" % (part.name, result["exception"]))
                return None

            self.ctx.stats_parts_instantiated += 1

            return result["shape"]



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

import time

from .ai_google import AiGoogle
from .ai_openai import AiOpenAI

from . import logging as pc_logging
from .user_config import user_config


models = [
    "gpt-3.5-turbo",
    "gpt-4",
    "gpt-4-vision-preview",
    "gemini-pro",
    "gemini-pro-vision",
]


class Ai(AiGoogle, AiOpenAI):
    def generate(
        self,
        action: str,
        package: str,
        item: str,
        prompt: str,
        config: dict[str, str],
        num_options: int = 1,
        image_filenames: list[str] = [],
    ):
        with pc_logging.Action("Ai" + action, package, item):
            # Determine the model to use
            if "model" in config and config["model"] is not None:
                model = config["model"]
            else:
                provider = config.get("provider", None)
                if provider is None:
                    if not user_config.openai_api_key is None:
                        provider = "openai"
                    else:
                        provider = "google"

                if provider == "google":
                    if len(image_filenames) > 0:
                        model = "gemini-pro-vision"
                    else:
                        model = "gemini-pro"
                elif provider == "openai":
                    if len(image_filenames) > 0:
                        model = "gpt-4-vision-preview"
                    else:
                        model = "gpt-4"
                else:
                    error = "Provider %s is not supported" % provider
                    pc_logging.error(error)
                    return []

            # Generate the content
            if not model in models:
                error = "Model %s is not supported" % model
                pc_logging.error(error)
                return []

            result = []
            if model.startswith("gemini"):
                try:
                    result = self.generate_google(
                        model,
                        prompt,
                        image_filenames,
                        config,
                        num_options,
                    )
                except Exception as e:
                    pc_logging.error(
                        "Failed to generate with Google: %s" % str(e)
                    )
                    time.sleep(1)  # Safeguard against exceeding quota

            elif model.startswith("gpt"):
                try:
                    result = self.generate_openai(
                        model,
                        prompt,
                        image_filenames,
                        config,
                        num_options,
                    )
                except Exception as e:
                    pc_logging.error(
                        "Failed to generate with OpenAI: %s" % str(e)
                    )
                    time.sleep(1)  # Safeguard against exceeding quota

            else:
                pc_logging.error(
                    "Failed to associate the model %s with the provider" % model
                )

            if result is None:
                return []
            return result



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

import asyncio

import copy
import os
import tempfile
import threading

from .ai import Ai
from . import logging as pc_logging
from . import sync_threads as pc_thread
from .utils import total_size

NUM_ALTERNATIVES_GEOMETRIC_MODELING = 3
NUM_ALTERNATIVES_MODEL_GENERATION = 3


class PartFactoryFeatureAi(Ai):
    """Used by all part factories that generate parts with GenAI."""

    part_type: str
    script_type: str
    prompt_suffix: str

    def __init__(self, part_config, part_type, script_type, prompt_suffix=""):
        self.part_type = part_type
        self.script_type = script_type
        self.prompt_suffix = prompt_suffix

        prompt = part_config.get("desc", None)
        if prompt is None:
            error = "%s: Prompt is not set" % part_config["name"]
            pc_logging.error(error)
            raise Exception(error)

        self.ai_config = copy.deepcopy(part_config)
        if not "numGeometricModeling" in self.ai_config:
            self.ai_config["numGeometricModeling"] = (
                NUM_ALTERNATIVES_GEOMETRIC_MODELING
            )
        if not "numModelGeneration" in self.ai_config:
            self.ai_config["numModelGeneration"] = (
                NUM_ALTERNATIVES_MODEL_GENERATION
            )
        if not "tokens" in self.ai_config:
            self.ai_config["tokens"] = 2000

    def on_init_ai(self):
        """This method must be executed at the very end of the part factory
        constructor to finalize the AI initialization. At the time of the call
        self.part and self.instantiate must be already defined."""
        self.part.generate = lambda: self._create_file()

        # If uncommented out, this makes the package initialization
        # unaccceptably slow
        # if (
        #     not os.path.exists(self.part_config["path"])
        #     or os.path.getsize(self.part_config["path"]) == 0
        # ):
        #     self._create_file()
        self.instantiate_orig = self.instantiate
        self.instantiate = self._instantiate_ai

    async def _instantiate_ai(self, part):
        """This is a wrapper for the instantiate method that ensures that
        the part is (re)generated before the instantiation."""
        if (
            not os.path.exists(self.part_config["path"])
            or os.path.getsize(self.part_config["path"]) == 0
        ):
            self._create_file()

        return await self.instantiate_orig(part)

    def _create_file(self):
        """This method is called to (re)generate the part."""

        # Geometric modeling
        modeling_options = []

        # def modeling_task():
        #     modeling_options.extend(self._geometric_modeling())
        # threads = []
        # for _ in range(NUM_ALTERNATIVES_GEOMETRIC_MODELING):
        #     thread = threading.Thread(target=modeling_task)
        #     thread.start()
        #     threads.append(thread)
        # for thread in threads:
        #     thread.join()

        while len(modeling_options) < self.ai_config["numGeometricModeling"]:
            modeling_options.extend(self._geometric_modeling())
            pc_logging.info(
                "Generated %d geometric modeling candidates"
                % len(modeling_options)
            )

        # For each remaining geometric modeling option,
        # generate a model and render an image
        script_candidates = []
        candidate_id = 0
        for modeling_option_idx, modeling_option in enumerate(modeling_options):
            pc_logging.debug(
                "Generated the geometric modeling candidate %d: %s"
                % (modeling_option_idx, modeling_option)
            )

            # Generate the models using the required language
            scripts = self._generate_script(modeling_option)

            for script in scripts:
                pc_logging.debug(
                    "Generated the script candidate %d: %s"
                    % (candidate_id, script)
                )

                # Render an image of the model
                image_filename = self._render_image(script, candidate_id)

                # Check if the model was valid
                if image_filename is not None:
                    # Record the valid model and the image
                    script_candidates.append((image_filename, script))

                candidate_id += 1

            pc_logging.info(
                "So far %d valid script candidates" % len(script_candidates)
            )

        if len(script_candidates) == 0:
            pc_logging.error(
                "No valid script generated. Try changing the prompt."
            )
            return

        if len(script_candidates) == 1:
            script = script_candidates[0][1]
        else:
            # Compare the images and select the best one
            script = self.select_best_image(script_candidates)

        if not script is None:
            f = open(self.part_config["path"], "w")
            f.write(script)
            f.close()

    def _geometric_modeling(self):
        """This method generates geometric modeling options for the part."""

        prompt = (
            """You are an engineer performing geometric modeling of mechanical parts.
Given a short verbal description of a part,
you are creating a detailed description of the geometric shapes
required to reproduce that part.
Create a detailed listing of all geometric shapes and how they are
located against each other
(including dimensions, distances and offset in millimeters,
and angles in degrees),
to reproduce the part with the following description:
%s
"""
            % self.ai_config["desc"]
        )

        if "properties" in self.part_config:
            properties = "\n".join(
                [
                    "  %s: %s" % (k, v)
                    for k, v in self.part_config["properties"].items()
                ]
            )
            prompt += (
                """

The part is further described by the following properties:
%s
"""
                % properties
            )

        image_filenames = self.ai_config.get("images", [])
        if len(image_filenames) > 0:
            prompt += """
            
The part is further described by the attached images.
"""

        options = self.generate(
            "Geometric",
            self.project.name,
            self.name,
            prompt,
            self.ai_config,
            self.ai_config[
                "numGeometricModeling"
            ],  # NUM_ALTERNATIVES_GEOMETRIC_MODELING,
            image_filenames=image_filenames,
        )
        return options

    def _generate_script(self, geometric_modeling):
        """This method generates a script given specific geometric modeling."""

        prompt = """Generate a %s to define a 3D model of a part defined by
the following geometric modeling:
        %s
        """ % (
            self.script_type,
            geometric_modeling,
        )

        image_filenames = self.ai_config.get("images", [])
        if len(image_filenames) > 0:
            prompt += """
            
The part is further described by the attached images.
"""

        prompt += """%s

IMPORTANT: Output the %s itself and do not add any text or comments before or after it.
""" % (
            self.prompt_suffix,
            self.script_type,
        )

        scripts = self.generate(
            "Script",
            self.project.name,
            self.name,
            prompt,
            self.ai_config,
            self.ai_config["numModelGeneration"],
            image_filenames=image_filenames,
        )

        # Sanitize the output to remove the decorations
        scripts = list(map(lambda s: self._sanitize_script(s), scripts))

        return scripts

    def _sanitize_script(self, script):
        """Cleans up the GenAI output to keep the code only."""

        # Remove code blocks
        script = "\n".join(
            list(
                filter(
                    lambda l: (
                        False
                        if l.startswith("```") or l.startswith(" ```")
                        else True
                    ),
                    script.split("\n"),
                )
            )
        )

        return script

    def _render_image(self, script, candidate_id):
        """This method ensures the validity of a script candidate by attempting to render it."""

        source_path = tempfile.mktemp(suffix=self.extension)
        with open(source_path, "w") as f:
            f.write(script)
            f.close()
        pc_logging.debug("Generated script: %s" % source_path)

        output_path = tempfile.mktemp(suffix=".png")
        pc_logging.debug("Attempting to render: %s" % output_path)

        # Create an ephemeral part for the given script candidate
        part = self._create_part(
            {
                "name": self.name + "_candidate_" + str(candidate_id),
                # "path": source_path, # TODO(clairbee): doesn't this work? why?
                "type": self.part_type,
                "mute": True,  # Suppress errors
            }
        )
        # Since this is an ephemeral part, certain things need to be tweaked
        part.instantiate = self.instantiate_orig  # Remove the AI wrapper
        part.path = source_path  # Set the path to the temporary script file
        pc_logging.debug("Part created: %.2f KB" % (total_size(part) / 1024.0))

        def render(part):
            try:
                coro = part.get_shape()
                with pc_logging.Action(
                    "Instantiate", part.project_name, part.name
                ):
                    shape = asyncio.run(coro)
                if not shape is None:
                    part.render_png(self.ctx, None, output_path)
            except Exception as e:
                part.error("Failed to render the image: %s" % e)

        try:
            # Given we don't know whether the current thread is already
            # running an event loop or not, we need to create a new thread
            t = threading.Thread(target=render, args=[part])
            t.start()
            t.join()
        except Exception as e:
            part.error("Failed to render the image: %s" % e)

        if len(part.errors) > 0:
            pc_logging.debug(
                "There were errors while attemtping to render the image"
            )
            pc_logging.debug("%s" % part.errors)

        os.unlink(source_path)
        if not os.path.exists(output_path) or os.path.getsize(output_path) == 0:
            pc_logging.info(
                "Script candidate %d: failed to render the image" % candidate_id
            )
            return None
        pc_logging.info(
            "Script candidate %d: successfully rendered the image"
            % candidate_id
        )
        return output_path

    def select_best_image(self, script_candidates):
        """Iterate over script_candidates and compare images.
        Each script candidate is [<image_filename>, <script>]."""

        if len(script_candidates) == 0:
            return None

        image_filenames = list(map(lambda c: c[0], script_candidates))

        prompt = (
            """
From the attached images, select the image that is the best fit
for the following description:
%s

Respond with the numeric index (starting with 1) of the best fit image.
No other text is acceptable.
Just the number.
"""
            % self.ai_config["desc"]
        )

        # Ask AI to compare the images
        responses = self.generate(
            "Compare",
            self.project.name,
            self.name,
            prompt,
            self.ai_config,
            1,
            image_filenames=image_filenames,
        )

        pc_logging.debug("Image comparison responses: %s" % responses)

        # Parse the response to get the number out
        for response in responses:
            words = response.split(" ")
            words = list(filter(lambda w: w.strip().isnumeric(), words))
            if len(words) == 1:
                idx = int(words[0])
                if idx > 0 and idx <= len(script_candidates):
                    pc_logging.info("Selected the best fit image: %d" % idx)
                    return script_candidates[idx - 1][1]

        return script_candidates[0][1]
